require 5.004;
use strict;

package HTML::FromText;
use Exporter;
use Text::Tabs 'expand';
use vars qw($RCSID $VERSION @EXPORT @ISA);
@ISA = qw(Exporter);
@EXPORT = qw(text2html);
$RCSID = q$Id: FromText.pm,v 1.11 1999/04/27 14:20:00 garethr Exp $;
$VERSION = '1.003';

# This list of protocols is taken from RFC 1630: "Universal Resource
# Identifiers in WWW".  The protocol "file" is omitted because experience
# suggests that it results in many false positives; "https" postdates RFC 1630.

my $protocol = join '|',
  qw(afs cid ftp gopher http https mid news nntp prospero telnet wais);

# The regular expressions matching e-mail addresses use the following syntax
# elements from RFC 822.  Because HTML::FromText has to spot e-mail addresses,
# not just verify them, I can't use the full details of structured field bodies
# (see http://www.perl.com/CPAN/authors/Tom_Christiansen/scripts/ckaddr.gz for
# more robust verification of e-mail addresses).
#
#   addr-spec   =  local-part "@" domain
#   local-part  =  word *("." word)
#   word        =  atom
#   domain      =  sub-domain *("." sub-domain)
#   sub-domain  =  domain-ref
#   domain-ref  =  atom
#   atom        =  1*<any CHAR except specials, SPACE and CTLs>
#   specials    =  "(" / ")" / "<" / ">" / "@" /  "," / ";" / ":" / "\"
#   		   / <"> /  "." / "[" / "]"
#
# I have ignored quoting, domain literals and comments.
#
# Note that '&' can legally appear in email addresses (for example,
# 'fred&barney@stonehenge.com').  If the 'metachars' option is passed to
# text2html then I must use '&amp;' to recognize '&'.  Thus the regular
# expression $atom[0] recognizes an atom in the case where the option
# 'metachars' is false; $atom[1] recognizes an atom in the case where
# 'metachars' is true.  Similarly for the regular expressions $email[0] and
# $email[1], which recognize e-mail addresses.

my @atom =
  ( '[!#$%&\'*+\\-/0123456789=?ABCDEFGHIJKLMNOPQRSTUVWXYZ^_`abcdefghijklmnopqrstuvwxyz{|}~]+',
    '(?:&amp;|[!#$%\'*+\\-/0123456789=?ABCDEFGHIJKLMNOPQRSTUVWXYZ^_`abcdefghijklmnopqrstuvwxyz{|}~])+' );

my @email = ( "$atom[0](\\.$atom[0])*\@$atom[0](\\.$atom[0])*",
	      "$atom[1](\\.$atom[1])*\@$atom[1](\\.$atom[1])*" );

sub text2html {
  local $_ = shift;		# Take a copy; don't modify in-place.
  return $_ unless $_;

  my %options = ( metachars => 1, @_ );

  if ($options{metachars}) {
    s/&/&amp;/g;
    s/</&lt;/g;
    s/>/&gt;/g;
    s/\"/&quot;/g;
  }

  # Spot electronic mail addresses and turn them into links.  Note that if the
  # `email' option is not set, then we ignore addresses unless prefixed by
  # `mailto:'.
  if ($options{email} or $options{urls}) {
    s|((?:mailto:)?)($email[$options{metachars}?1:0])|($options{email} or $1) ? "<TT><A HREF=\"mailto:$2\">$1$2</A></TT>" : $2|ge;
  }

  if ($options{urls}) {
    s|\b((?:$protocol):\S+[\w/])|<TT><A HREF="$1">$1</A></TT>|g;
  }

  if ($options{bold}) {
    s#(^|\W)\*(\w+)\*(\W|$)#$1<B>$2</B>$3#g;
  }

  if ($options{underline}) {
    s#(^|\W)_(\w+)_(\W|$)#$1<U>$2</U>$3#g;
  }

  if ($options{pre}) {
    $_ = "<PRE>" . $_ . "</PRE>";
  } elsif ($options{lines}) {
    my @lines = split /\r?\n/;
    if ($options{spaces}) {
      @lines = expand(@lines);
      foreach (@lines) {
	s/ /&nbsp;/g;
      }
    }
    $_ = join "<BR>\n", @lines;
  } elsif ($options{paras}) {
    my @paras;

    # Remove initial and final blank lines.
    s/^(?:\s*?\n)+//;
    s/(?:\n\s*?)+$//;

    # Split on a different regexp depending on what kinds of paragraphs will be
    # recognised later.  The idea is that bulleted lists like this:
    #
    #     * item 1
    #     * item 2
    #
    # will be recognised as multiple paragraphs if the 'bullets' option is
    # supplied, but as a single paragraph otherwise.  (Similarly for numbered
    # lists).
    if ($options{bullets} and $options{numbers}) {
      @paras = split
	/(?:\s*\n)+                  # (0 or more blank lines, followed by LF)
         (?:\s*\n                    # Either 1 or more blank lines, or
          |(?=\s*[*-]\s+             #   bulleted item follows, or
            |\s*(?:\d+)[.\)\]]?\s+)) #   numbered item follows
        /x;
    } elsif ($options{bullets}) {
      @paras = split
	/(?:\s*\n)+                  # (0 or more blank lines, followed by LF)
         (?:\s*\n                    # Either 1 or more blank lines, or
          |(?=\s*[*-]\s+))           #   bulleted item follows.
        /x;
    } elsif ($options{numbers}) {
      @paras = split
	/(?:\s*\n)+                  # (0 or more blank lines, followed by LF)
         (?:\s*\n                    # Either 1 or more blank lines, or
         |(?=\s*(?:\d+)[.\)\]]?\s+)) #   numbered item follows.
        /x;
    } else {
      @paras = split 
	/\s*\n(?:\s*\n)+	     # 1 or more blank lines.
        /x;
    }
    my $last = '';		# List type (OL/UL) of last paragraph
    my $this;			# List type (OL/UL) of this paragraph
    my $first = 1;		# True if this is first paragraph
    foreach (@paras) {
      $this = '';
      if ($options{title} and $first) {
	s|^|<H1>|;
	s|$|</H1>|;
      } elsif ($options{headings} and /^(\d+(\.\d+)*)\.?\s/) {
	my $number = $1;
	my $level = 1 + ($number =~ tr/././);
	$level = 1 if $level < 1;
	$level = 6 if $level > 6;
	s|^|<H$level>|;
	s|$|</H$level>|;
      } elsif ($options{bullets} and s/^\s*[*-]\s+/<LI><P>/) {
	s|$|</P>|;
	$this = 'UL';
      } elsif ($options{numbers}
	       and s/^\s*(\d+)[.\)\]]?\s+/<LI VALUE="$1"><P>/) {
	s|$|</P>|;
	$this = 'OL';
      } elsif ($options{blockquotes} and /^(\s+).*(?:\n\1.*)*$/) {
	# Every line in the paragraph starts with at white space, the common
	# whitespace being in $1.  Remove the common initial whitespace,
	s/^$1//gm;

	# Split into lines, expand tabs, and join back with <BR>.
	my @lines = split /\r?\n/;
	@lines = expand(@lines);
	$_ = join '','<BLOCKQUOTE>',(join "<BR>\n", @lines),'</BLOCKQUOTE>';
      } else {
	s|^|<P>|;
	s|$|</P>|;
      }
      if ($this ne $last) {
	s|^|<$this>| if ($this ne '');
	s|^|</$last>| if ($last ne '');
      }
      $last = $this;
      $first = 0;
    }
    if ($this ne '') {
      push @paras, "</$this>";
    }
    $_ = join "\n", @paras;
  }
  return $_;
}

1;
